#include <stdio.h>  
#include <stdlib.h>
#include <math.h> 
#include <string.h>

typedef struct {
  int    ord;
  double crd[2];
} NODE;

typedef struct {
  int    ord;
  int    vtx[3];
  double center[2];
} TRIG;

NODE   *nodes;
TRIG   *trigs;
int    n_nodes,n_trigs;

double df(double pt[]);


/*****Density fuction*****/
double get_Density(double pt[])
{
  double  dist;

  dist = sqrt(pt[0]*pt[0]+pt[1]*pt[1]);

  return exp(-10.0*fabs(dist-0.5));
}


/*****Read the nodes information from the input file "nodes.dat"*****/
void read_Nodes_From_File()
{
  FILE*  fp;
  double ftemp;
  int    i,j,ntemp;
                         
  fp=fopen("nodes.dat","rt");
  fscanf(fp,"%d",&n_nodes);
  for (i=0;i<3;i++) fscanf(fp,"%d",&ntemp);
  nodes = (NODE*)calloc(n_nodes,sizeof(NODE));
  for (i=0;i<n_nodes;i++) {
      fscanf(fp,"%d",&ntemp);
      (nodes+i)->ord = ntemp-1;
      for (j=0;j<2;j++) {
          fscanf(fp,"%lf",&ftemp);
          (nodes+i)->crd[j] = ftemp;
      }
      fscanf(fp,"%d",&ntemp);
  }                 
  fclose(fp);      
}


/*****Read the triangles information from the input file "trigs.dat"*****/
void read_Trigs_From_File()
{
  FILE*  fp;
  int    i,j,k,ntemp;
                                                   
  fp=fopen("trigs.dat","rt");   
  fscanf(fp,"%d",&n_trigs);
  for (i=0;i<2;i++) fscanf(fp,"%d",&ntemp);
  trigs = (TRIG*)calloc(n_trigs,sizeof(TRIG));
  for (i=0;i<n_trigs;i++) {
      fscanf(fp,"%d",&ntemp);
      (trigs+i)->ord = ntemp-1;
      for (j=0;j<3;j++) {
          fscanf(fp,"%d",&k);
          (trigs+i)->vtx[j] = k-1;
      }
  }
  fclose(fp);
}


/*****Write the density information files for density function*****/
void write_Dens_Info()
{
  FILE*  fp;
  int    i,j,k,nj;
  double fd,weig;
              
  fp=fopen("dens_nodes.dat","wt");
  weig = 1.0/3.0;
  for (i=0;i<n_trigs;i++) {
      for (k=0;k<2;k++) (trigs+i)->center[k] = 0.0;
      for (j=0;j<3;j++) {
          nj = (trigs+i)->vtx[j];
          for (k=0;k<2;k++)
              (trigs+i)->center[k] = (trigs+i)->center[k]+weig*(nodes+nj)->crd[k];
      }
      for (k=0;k<2;k++) fprintf(fp,"%.10f  ",(trigs+i)->center[k]);
      fprintf(fp,"\n");
  }
  fclose(fp);
  fp=fopen("dens_valus.dat","wt");
  for (i=0;i<n_trigs;i++) { 
      fd = get_Density((trigs+i)->center);   
      fprintf(fp,"%20.10f ",fd);
      fprintf(fp,"\n");
  }
  fclose(fp);
}



int main()
{
  read_Nodes_From_File();
  read_Trigs_From_File();
  printf("# of background nodes = %7d\n",n_trigs);
  printf("Generate dens_nodes.dat and dens_valus.dat ...\n");
  write_Dens_Info();
}
