/*
 * This program convert a 2D tetrahedron mesh to 3D prism mesh,
 * by make tensor product in z direction.
 *
 * Usage: ./tri2prism tria.mesh 
 *
 * By Leng Wei, LSEC, Jan 31 2010.
 * 
 *  */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <phg.h>

#define NUMBERING_Z 0

/******************/
/* GEO DEFINATION */
/******************/
#define phgElemInfoCount 6

typedef enum {
    QUADRILATERAL = 0,
    TRIANGLE	= 1,
    TETRA	= 2,
    PYRAMID	= 3,
    PRISM	= 4,
    BRICK	= 5
} ELEM_TYPE;

typedef short EDGE_INFO[2];	/* list of vertices */
typedef short FACE_INFO[5];	/* 0: # of vertices, 1..: list of vertices */
typedef void (*ADD_TETRA_FUNC)(int v0, int v1, int v2, int v3, int bound_type[]);
typedef void (*FUNC2TET)(int verts[], ADD_TETRA_FUNC func, int bound_type[]);

typedef struct {
    const char  *name;		/* name of the element type */
    EDGE_INFO   *edge_info;	/* list of edges */
    FACE_INFO   *face_info;	/* list of faces */
    short       nvert;		/* number of vertices */
    short	nedge;		/* number of edges */
    short       nface;		/* number of faces */
    FUNC2TET    func2tet;	/* function to covert to tet */
} ELEM_INFO;

ELEM_INFO phgElemInfo[] = {
    /* name	list of edges	list of faces	nvert	nedge	nface  func2tet*/
    {"quad",	NULL,	NULL,	4,	4,	4, NULL},
    {"tria",	NULL,	NULL,	3,	3,	3, NULL},
};

int prism_face[] = {1, 2, 0};

/*****************/
/* Reading Utils */
/*****************/
#define phgWarning(...) fprintf(stderr, __VA_ARGS__)
#define phgError(verb, ...) fprintf(stderr, __VA_ARGS__)
#define phgInfo(verb, ...) fprintf(stdout, __VA_ARGS__)
#define phgCalloc calloc
#define phgRealloc realloc

#define READ_NUMBER						\
    if (!get_token(fp, token)) strcpy(token, "End ALL");	\
    if (isalpha((int)(token[0]))) {				\
	phgWarning("fewer entries (%d) than expected.\n", i);	\
	abort();						\
	break;							\
    }

#define GET_TOKEN {				\
	if (!get_token(fp, token))		\
	    goto error;				\
    }

#define UNUSED_LINE { 				\
	fgets(token, 1024, fp);			\
	fprintf(stderr, "Unuesd: %s", token);	\
    }
#define NEXT_LINE UNUSED_LINE

#define UNUSED_LINE_CHECK(str) {				\
	fgets(token, 1024, fp);					\
	if (!strcmp(token, str))				\
	    phgError(-1, "Unmatched phase when read mesh file,"	\
		     " line:%d!\n", __LINE__);			\
	fprintf(stderr, "Unuesd: %s", token);			\
    }


/* 
 * Array abbr.
 *  
 *  */
#define Neigh(i, j) neigh[i*3 + j]
#define Edges(i, j) edges[i*2 + j]
#define OFFSET 0



static BOOLEAN
get_token(FILE *fp, char *token)
{
    int c;
    char *p;

    while (TRUE) {
	if (fscanf(fp, "%s", token) != 1)
	    return FALSE;
	if (token[0] != '#')
	    break;
	/* skip to newline */
	do
	    if ((c = fgetc(fp)) == EOF)
		return FALSE;
	while (c != '\n');
    }
    if ((p = strchr(token, '#')) != NULL)
	*p = '\0';
    return TRUE;
}














int 
main(int argc, char *argv[])
{
    FILE *fp;
    GRID *g;
    char token[1024]; 
    INT i, j, n, k, Nly;
    INT numnp, nelem, ngrps, nbdry, *neigh, *edges;
    FLOAT *R;

    g = calloc(1, sizeof(*g));
    
    /***********************/
    /* Control Information */
    /***********************/
    if (0) {
    error:
	phgError(1, "invalid or unsupported input file, abort.\n");
    }


    /*********************/
    /* Nodal Coordinates */
    /*********************/
    fp = fopen("nodes.dat", "r");
    assert(fp != NULL);
    GET_TOKEN; numnp = atoi(token);
    GET_TOKEN; GET_TOKEN; GET_TOKEN; 
    n = numnp;
    phgInfo(0, "number of vertices: %d\n", n);
    g->verts = calloc(n, sizeof(*g->verts));
    for (i = 0; i < n; i++) {
	READ_NUMBER;
	GET_TOKEN; g->verts[i][0] = atof(token);
	GET_TOKEN; g->verts[i][1] = atof(token);
	READ_NUMBER;
    }
    g->nvert = n;
    UNUSED_LINE_CHECK("# end\n")
	fclose(fp);


    /*****************************/
    /* Element/Cell Connectivity */
    /*****************************/
    fp = fopen("trigs.dat", "r");
    assert(fp != NULL);
    GET_TOKEN; nelem = atoi(token);
    GET_TOKEN; GET_TOKEN;
    n = nelem;
    phgInfo(0, "number of triangles: %d\n", n);
    g->roots = phgCalloc(n, sizeof(*g->roots));
    for (i = 0; i < n; i++) {
	READ_NUMBER;
	/* c convention */
	GET_TOKEN; g->roots[i].verts[0] = atof(token) -1;
	GET_TOKEN; g->roots[i].verts[1] = atof(token) -1;
	GET_TOKEN; g->roots[i].verts[2] = atof(token) -1;
    }
    UNUSED_LINE_CHECK("# end\n")
	fclose(fp);


    /****************************/
    /* Boundary Conditions Sets */
    /****************************/
    fp = fopen("neigs.dat", "r");
    assert(fp != NULL);
    nbdry = 0; 
    neigh = phgCalloc(nelem * 3, sizeof(*neigh));
    for (i = 0; i < nelem; i++) {
	for (k = 0; k < 3; k++) {
	    GET_TOKEN; Neigh(i, k) = atoi(token);
        //printf("%d\n", Neigh(i,k));
	    if (Neigh(i, k) == -1)
		nbdry++;
	}
	//printf("%d %d %d\n", Neigh(i, 0), Neigh(i, 1), Neigh(i, 2));
    }
    phgInfo(0, "number of bdry edges %d\n", nbdry);
    UNUSED_LINE_CHECK("# end\n");
	fclose(fp);
    
    edges = phgCalloc(nbdry * 2, sizeof(*neigh));
    nbdry = 0;
    for (i = 0; i < nelem; i++) 
	for (k = 0; k < 3; k++) 
	    if (Neigh(i, k) == -1) {
		Edges(nbdry, 0) = i;
		Edges(nbdry, 1) = k;
		nbdry++;
	    }


    /**************/
    /* Layer Info */
    /**************/
    //"layers.dat"
    fp = fopen(argv[1], "r");
    assert(fp != NULL);
    fscanf(fp, "%d", &Nly);
    fprintf(stderr, "Layer: %d\n", Nly);
    R = calloc(Nly+1, sizeof(*R));
    FLOAT dR = 1. / Nly;
    for (i = 0; i < Nly+1; i++) {
	fscanf(fp, "%lf", R + i);
	fprintf(stderr, "   L[%2d]: %f\n", i, R[i]);
	//R[i] = i * dR;
    }
    //UNUSED_LINE_CHECK("# end\n");
    fclose(fp);



















    /****************************/
    /* Build 3D mesh and output */
    /****************************/
    fp = fopen("prism.neu", "w");
    assert(fp != NULL);

    ngrps = 1;

    /* head */
    fprintf(fp, "%20s\n%20s\n%20s\n%20s\n%20s\n%20s\n",
	    "        CONTROL INFO 2.3.16",
	    "** GAMBIT NEUTRAL FILE",
	    "default_id",
	    "PROGRAM:                Gambit     VERSION:  2.3.16",
	    " 7 Jan 2010    10:44:42",
	    "     NUMNP     NELEM     NGRPS    NBSETS     NDFCD     NDFVL"
	    );
    fprintf(fp, "%10d%10d%10d%10d%10d%10d\n", 
	    numnp * (Nly + 1), nelem * Nly, ngrps, 
	    3, 3, 3);
    fprintf(fp, "ENDOFSECTION\n");

    /* node */
    fprintf(fp, "%20s\n","    NODAL COORDINATES 2.3.16");
#if NUMBERING_Z
    for (k = 0; k < Nly+1; k++) {
	for (i = 0; i < numnp; i++) {
	    fprintf(fp, "%10d%20.12e%20.12e%20.12e\n", 
		    k*numnp + i + 1, 
		    g->verts[i][0], 
		    g->verts[i][1], R[k]);
	}
    }
#else
    for (i = 0; i < numnp; i++) {
	for (k = 0; k < Nly+1; k++) {
	    fprintf(fp, "%10d%20.12e%20.12e%20.12e\n", 
		    k + i*(Nly+1) +1, 
		    g->verts[i][0], 
		    g->verts[i][1], R[k]);
	}
    }
#endif
    fprintf(fp, "ENDOFSECTION\n");


    /* Vertex Z connection */
    FILE *fp_vert = fopen("node.Z", "w");
#if NUMBERING_Z
    abort();
#else
    for (i = 0; i < numnp; i++) {
	fprintf(fp_vert, "%d %d ", i+1, Nly+1);
	for (k = 0; k < Nly+1; k++) {
	    fprintf(fp_vert, "%d ", k + i*(Nly+1) +1);
	}
	fprintf(fp_vert, "\n");
    }
#endif
    fclose(fp_vert);


    /* elements */
    fprintf(fp, "%20s\n","      ELEMENTS/CELLS 2.3.16");
#if NUMBERING_Z
    for (k = 0; k < Nly; k++) {
	for (i = 0; i < nelem; i++) {
	    int v[3];
	    for (j = 0; j < 3; j++)
		v[j] = g->roots[i].verts[j];

	    fprintf(fp, "%8d %2d %2d %8d %8d %8d %8d %8d %8d\n", 
		    k*nelem + i +1, 5, 6, 
		    v[0] + (k)*numnp +1, 
		    v[1] + (k)*numnp +1, 
		    v[2] + (k)*numnp +1, 
		    v[0] + (k+1)*numnp +1, 
		    v[1] + (k+1)*numnp +1, 
		    v[2] + (k+1)*numnp +1);
	}
    }
#else
    for (i = 0; i < nelem; i++) {
	int v[3];
	for (j = 0; j < 3; j++)
	    v[j] = g->roots[i].verts[j];

	for (k = 0; k < Nly; k++) {
	    fprintf(fp, "%8d %2d %2d %8d %8d %8d %8d %8d %8d\n", 
		    k + i*Nly +1, 5, 6, 
		    v[0]*(Nly+1) + (k) +1, 
		    v[1]*(Nly+1) + (k) +1, 
		    v[2]*(Nly+1) + (k) +1, 
		    v[0]*(Nly+1) + (k+1) +1, 
		    v[1]*(Nly+1) + (k+1) +1, 
		    v[2]*(Nly+1) + (k+1) +1);
	}
    }
#endif
    fprintf(fp, "ENDOFSECTION\n");

    /* groups */
    fprintf(fp, "	    ELEMENT GROUP 2.3.1\n"
	    "GROUP:          1 ELEMENTS:%10d MATERIAL:          2 NFLAGS:          1\n"
	    "                           fluid\n", nelem * Nly);
    fprintf(fp, "%8d\n", 0);
    for (k = 0; k < Nly; k++)
	for (i = 0; i < nelem; i++) {
	    fprintf(fp, "%8d", k*nelem + i +1);
	    if ((k*nelem + i +1) % 10 == 0
		&& !(k == Nly-1 && i == nelem-1)) /* last one need not line break */
		fprintf(fp, "\n");
	}
    fprintf(fp, "\nENDOFSECTION\n");


    /* boundary conditions */
    fprintf(stderr, "# of Boundary edges: %d\n", 
	    nbdry);















    /* bottom */
    fprintf(fp, " BOUNDARY CONDITIONS 2.3.16\n");
    fprintf(fp, "                  bottom");
    fprintf(fp, "%8d%8d%8d%8d\n", 1, nelem, 0, 6);
#if NUMBERING_Z
    for (i = 0; i < nelem; i++) {
	fprintf(fp, "%10d%5d%5d\n", 
		i +1, 5, 4);
    }    
#else
    for (i = 0; i < nelem; i++) {
	fprintf(fp, "%10d%5d%5d\n", 
		i*Nly +1, 5, 4);
    }    
#endif
    fprintf(fp, "ENDOFSECTION\n");












    /* top */
    fprintf(fp, " BOUNDARY CONDITIONS 2.3.16\n");
    fprintf(fp, "                  top");
    fprintf(fp, "%8d%8d%8d%8d\n", 1, nelem, 0, 6);
#if NUMBERING_Z
    for (i = 0; i < nelem; i++) {
	fprintf(fp, "%10d%5d%5d\n", 
		i + (Nly-1)*nelem +1, 5, 5);
    }    
#else
    for (i = 0; i < nelem; i++) {
	fprintf(fp, "%10d%5d%5d\n", 
		i*Nly + Nly-1 +1, 5, 5);
    }    
#endif
    fprintf(fp, "ENDOFSECTION\n");











    /* lateral */
    fprintf(fp, " BOUNDARY CONDITIONS 2.3.16\n");
    fprintf(fp, "                  lateral");
    fprintf(fp, "%8d%8d%8d%8d\n", 1, nbdry * Nly, 0, 6);

#if NUMBERING_Z
    for (i = 0; i < nbdry; i++) {
	for (k = 0; k < Nly; k++)
	    fprintf(fp, "%10d%5d%5d\n", 
		    Edges(i, 0) + k * nelem +1, 5, 
		    prism_face[Edges(i, 1)] +1);
    }
#else
    for (i = 0; i < nbdry; i++) {
	for (k = 0; k < Nly; k++)
	    fprintf(fp, "%10d%5d%5d\n", 
		    Edges(i, 0)*Nly + k +1, 5, 
		    prism_face[Edges(i, 1)] +1);
    }
    /* for (i = 0; i < n; i++, listf++) { */
    /* 	for (k = 0; k < Nly; k++) */
    /* 	    fprintf(fp, "%10d%5d%5d\n",  */
    /* 		    (*listf)[0]*Nly + k +1, 5,  */
    /* 		    (*listf)[1] +1); */
    /* } */
#endif
    fprintf(fp, "ENDOFSECTION\n");





    fclose(fp);
    return 0;
}
