#include "phg.h"
#include <stdlib.h>
#include <string.h>
#include <strings.h>	/* bzero() */
#include <math.h>
#include <stdarg.h>

#define SQUARE(x) ((x)*(x))
#define INNER_PRODUCT(p, q)			\
    (*(p    ) * *(q    ) +			\
     *(p + 1) * *(q + 1) +			\
     *(p + 2) * *(q + 2))
#define MAX(a, b) ((a) > (b) ? (a) : (b))
#define MIN(a, b) ((a) < (b) ? (a) : (b))

#define BC_TOP BDRY_USER1
#define BC_BOT BDRY_USER2
#define BC_BOT_S BDRY_USER6
#define BC_LAT BDRY_USER3
#define BC_HEA BDRY_USER4
#define BC_TER BDRY_USER5
#define BC_BOT_SLIP BDRY_USER7

# define n  3
# define rho_i 910
# define rho_w 1000
# define gravity 9.8
# define sec_per_y 31536000.0
# define visc_init 1.0e7
# define beta_slip 1e4

void func_bed_z(FLOAT x, FLOAT y, FLOAT z, FLOAT *bed_z)
{
    FLOAT bed;
    //*bed_z = 720 - 1.038e-3*x;
    *bed_z = -100 - x/1000.0;
}

void
func_ice_shelf_mask(FLOAT x, FLOAT y, FLOAT z, FLOAT *ice_shelf_mask)
{
    //double **data_bed, **data_sur, **data_thk;
    x *= 1000;
    y *= 1000;    
    //bed = -100-x/1000.0;

    if (x>5.4e5)
        ice_shelf_mask[0] = 1;
    else
        ice_shelf_mask[0] = 0;
}

void
func_ice_domain(FLOAT x, FLOAT y, FLOAT z, FLOAT *coord)
{
    FLOAT sur;
    FLOAT thk;
    FLOAT bed;

    x *= 1000; 
    y *= 1000;

    coord[0] = x;
    coord[1] = y;

    FLOAT bed_g, thk_g, bot;

    FLOAT SPY = 365*3600*24;
    FLOAT m = 1.0/3.;
    //INT n = 3;
    FLOAT rho1 = 910.0, rho1_w = 1.0e3;
    FLOAT g = 9.8;
    FLOAT a = 0.5/SPY;
    //FLOAT C = 1e3*SPY;
    //FLOAT C = 7.624e6;
    FLOAT C = 1.0e7;

    FLOAT x_g = 5.4e5;
    FLOAT u_g = a*x_g;
    FLOAT r = rho1/rho1_w;
    FLOAT A = 1e-25;

    func_bed_z(x, y, z, &bed);

    func_bed_z(x_g, y, z, &bed_g);

    FLOAT h_g = rho1_w/rho1*fabs(bed_g);

    if (x < x_g)
    {
        thk = pow(pow(h_g, m+2) + (m+2)/(m+1)*C*pow(a,m)/rho1/g*(pow(x_g, m+1) - pow(x,m+1)),1.0/(m+2));
        //*data  = bed + z*thk;
        bot = bed;
        sur = bed + thk;
    }
    else
    {
        
        thk = h_g*(u_g + a*(x-x_g))/pow(pow(u_g,3+1) + pow(h_g,3+1)*pow((1-r)*rho1*g/4,3)*A*(pow((u_g + a*(x-x_g)),3+1) - pow(u_g,3+1))/a,1.0/(3+1));
        //*data = (1-r)*thk - (1-z)*thk;
        bot = (1-r)*thk - thk;
        sur = (1-r)*thk;

    }
    coord[2] = bot + z*thk;
}

void build_ice_domain(GRID *g)
{
    SIMPLEX *e;
    DOF *coord;
    int s, i;
    FLOAT eps = 1.0e-8;
    
    // For MISMIP2D
    //grid_data = read_txt_data_1D("grid.txt"); 
    DOF *ice_shelf_mask = phgDofNew(g, DOF_P3, 1, "ice_shelf_mask", func_ice_shelf_mask);
    coord = phgDofNew(g, DOF_P1, Dim, "coord", func_ice_domain);
    phgExportVTK(g, "ice_shelf_mask.vtk", ice_shelf_mask, NULL);


    ForAllElements(g, e)
    {
        for (s=0; s<NFace; s++)
        {
            FLOAT normal_top[] = {0, 0, 1}, normal_bot[] = {0, 0, -1};
            FLOAT normal_hea[] = {-1, 0, 0}, normal_ter[] = {1, 0, 0};
            const FLOAT *normal = phgGeomGetFaceOutNormal(g, e, s);
            
            //e->bound_type[s] &= ~DIRICHLET;

            if (e->neighbours[s] != NULL)
                continue;

            if (fabs(1-INNER_PRODUCT(normal, normal_top))<eps)
            {
                e->bound_type[s] = BC_TOP;
            }
            else if (fabs(1-INNER_PRODUCT(normal, normal_bot))<eps)
            {
                e->bound_type[s] = BC_BOT;
                if (*DofFaceData(ice_shelf_mask, e->faces[s])==1)
                {
                	e->bound_type[s] = BC_BOT_S;
                }
            }
            else if (fabs(1-INNER_PRODUCT(normal, normal_hea))<eps)
            {
            	e->bound_type[s] = BC_HEA;
            }
            else if (fabs(1-INNER_PRODUCT(normal, normal_ter))<eps)
            {
                e->bound_type[s] = BC_TER;
            }
            else
            {
                e->bound_type[s] = BC_LAT;
            }
        }
    }
    phgUpdateBoundaryTypes(g);

    ForAllElements(g, e) 
    {
	    for (i = 0; i < NVert; i++) 
        {
	        FLOAT *coord_old = g->verts[e->verts[i]];
            FLOAT *coord_new = DofVertexData(coord, e->verts[i]);
            memcpy(coord_old, coord_new, Dim*sizeof(*coord_old));
        }
    }

    phgGeomInit_(g, TRUE);

   phgDofFree(&coord);
   return;

}

void
packDof(DOF **dofs, int ndof, ...)
{
    int i;
    va_list ap;

    va_start(ap, ndof);
    for (i = 0; i < ndof; i++)
	dofs[i] = va_arg(ap, DOF *);
    va_end(ap);
    return;
}

void
unpackDof(DOF **dofs, int ndof, ...)
{
    int i;
    va_list ap;
    DOF **pdof;

    va_start(ap, ndof);
    for (i = 0; i < ndof; i++) {
	pdof = va_arg(ap, DOF **);
	*pdof = dofs[i];
    }
    va_end(ap);
    return;
}

void
packMat(MAT **mats, int nmat, ...)
{
    int i;
    va_list ap;

    va_start(ap, nmat);
    for (i = 0; i < nmat; i++)
	mats[i] = va_arg(ap, MAT *);
    va_end(ap);
    return;
}

void
unpackMat(MAT **mats, int nmat, ...)
{
    int i;
    va_list ap;
    MAT **pmat;

    va_start(ap, nmat);
    for (i = 0; i < nmat; i++) {
	pmat = va_arg(ap, MAT **);
	*pmat = mats[i];
    }
    va_end(ap);
    return;
}

FLOAT *get_grad_bas_product(const FLOAT *gbas_i, const FLOAT *gbas_j)
{
    static FLOAT prod[Dim][Dim];

    prod[0][0] = gbas_i[0] * gbas_j[0] + .5 * (gbas_i[1] * gbas_j[1] + gbas_i[2] * gbas_j[2]);
    prod[0][1] = .5 * gbas_i[0] * gbas_j[1];
    prod[0][2] = .5 * gbas_i[0] * gbas_j[2];

    prod[1][0] = .5 * gbas_i[1] * gbas_j[0]; 
    prod[1][1] = gbas_i[1] * gbas_j[1] + .5 * (gbas_i[0] * gbas_j[0] + gbas_i[2] * gbas_j[2]);
    prod[1][2] = .5 * gbas_i[1] * gbas_j[2];

    prod[2][0] = .5 * gbas_i[2] * gbas_j[0]; 
    prod[2][1] = .5 * gbas_i[2] * gbas_j[1]; 
    prod[2][2] = gbas_i[2] * gbas_j[2] + .5 * (gbas_i[0] * gbas_j[0] + gbas_i[1] * gbas_j[1]);

    return prod[0];
}

FLOAT *get_bas_dot_normal(const FLOAT *bas, const FLOAT *normal, INT i_e, INT j_e)
{
	static FLOAT basdotn[Dim][Dim];

	basdotn[0][0] = normal[0]*normal[0]*bas[i_e]*bas[j_e];
	basdotn[0][1] = normal[1]*normal[0]*bas[i_e]*bas[j_e];
	basdotn[0][2] = normal[2]*normal[0]*bas[i_e]*bas[j_e];
	basdotn[1][0] = normal[0]*normal[1]*bas[i_e]*bas[j_e];
	basdotn[1][1] = normal[1]*normal[1]*bas[i_e]*bas[j_e];
	basdotn[1][2] = normal[2]*normal[1]*bas[i_e]*bas[j_e];
	basdotn[2][0] = normal[0]*normal[2]*bas[i_e]*bas[j_e];
	basdotn[2][1] = normal[1]*normal[2]*bas[i_e]*bas[j_e];
	basdotn[2][2] = normal[2]*normal[2]*bas[i_e]*bas[j_e];

	return basdotn[0];
}

FLOAT get_frobenius_eps_u_from_grad_u(const FLOAT *a)
{
    static FLOAT c;
    c = SQUARE(a[0]) + SQUARE(a[4])*1 + SQUARE(a[8])*1 + 
        0.5*(SQUARE(a[1]+a[3])*1 + SQUARE(a[2]+a[6]*1) + SQUARE(a[5]+a[7])*1);
    return c;
}

FLOAT
get_viscosity(const FLOAT *grad_u_quad_q)
{
    static FLOAT strain_rate[Dim*Dim], strain_rate_e, viscosity;
    FLOAT A;
    //int n = 3;

    strain_rate[0] = grad_u_quad_q[0];
    strain_rate[1] = 0.5*(grad_u_quad_q[1]+grad_u_quad_q[3])*0;
    strain_rate[2] = 0.5*(grad_u_quad_q[2]+grad_u_quad_q[6]*0);
    strain_rate[3] = strain_rate[1];
    strain_rate[4] = grad_u_quad_q[4]*0;
    strain_rate[5] = 0.5*(grad_u_quad_q[5]+grad_u_quad_q[7])*0;
    strain_rate[6] = strain_rate[2];
    strain_rate[7] = strain_rate[5];
    strain_rate[8] = grad_u_quad_q[8]*0;


    strain_rate_e = sqrt(0.5*(SQUARE(strain_rate[0])
                              +SQUARE(strain_rate[4])
                              +SQUARE(strain_rate[8])
                              +2*SQUARE(strain_rate[1])
                              +2*SQUARE(strain_rate[2])
                              +2*SQUARE(strain_rate[5])
                              ));
    if (strain_rate_e > 1e8)
        printf("strain_rate_e: %e\n", strain_rate_e);

    if (strain_rate_e < 1.0e-15)
        strain_rate_e = 1.0e-15;


    A = 1.0e-25*sec_per_y;
    viscosity = 0.5*pow(A, -1.0/n)*pow(strain_rate_e, -1.0+1.0/n);

    return viscosity;
}

void
build_matrices(SOLVER *solver, DOF **dofs, MAT **mats)
{
    DOF *velocity = dofs[0], *pressure = dofs[1];
    DOF *velocity_grad;
    int M = velocity->type->nbas;	
    int N = pressure->type->nbas;	
    GRID *g = velocity->g;
    SIMPLEX *e;
    MAP *umap, *pmap;
    MAT *matF, *matB, *matBt, *matC;

    int i, j, k, l, q, quad_order, s;
    FLOAT *w, *pts;
    FLOAT gravity_acc[Dim] = {0, 0, gravity}, eps = 1.0e-10;
    FLOAT det, vol;
    //const FLOAT *grad_bas_u_i_q, *grad_bas_u_j_q, *bas_p_j;
    //FLOAT *grad_bas_product;
    QUAD *quad;
    quad_order = 5;
    quad = phgQuadGetQuad3D(quad_order);
    w = quad->weights;
    pts = quad->points;

    FLOAT F[M][Dim][M][Dim], B[N][M][Dim], Bt[M][Dim][N], C[N][N] , rhsu[M][Dim], bufu[M], bufp[N];
    INT Iu[M][Dim], Ju[Dim][M], Ip[N];

    unpackDof(dofs, 3, &velocity, &pressure, &velocity_grad);
    unpackMat(mats, 4, &matF, &matB, &matBt, &matC);


    ForAllElements(g, e)
    {
        memset(F, 0, sizeof F); memset(B, 0, sizeof B); memset(Bt, 0, sizeof Bt); memset(C, 0, sizeof C); 
        
        //const FLOAT *visc_quad = phgQuadGetDofValues(e, viscosity, quad);
        const FLOAT *velocity_grad_quad = phgQuadGetDofValues(e, velocity_grad, quad);
        //const FLOAT *temprature_quad = phgQuadGetDofValues(e, temperature, quad);

        for (i = 0; i < M; i++)
	    for (k = 0; k < Dim; k++)
		Ju[k][i] = Iu[i][k] = phgMapE2L(matF->cmap, 0, e, i * Dim + k);

        for (i = 0; i < N; i++)
	    Ip[i] = phgMapE2L(matC->cmap, 0, e, i);

        
        vol = phgGeomGetVolume(g, e);

        for (q = 0; q < quad->npoints; q++)
        {
            for (i = 0; i < M; i++)
            {
                const FLOAT *grad_bas_u_i_q = phgQuadGetBasisGradient(e, velocity, i, quad) + q*Dim;
            for (j = 0; j < M; j++)
            {
                const FLOAT *grad_bas_u_j_q = phgQuadGetBasisGradient(e, velocity, j, quad) + q*Dim;
                const FLOAT *grad_bas_product = get_grad_bas_product(grad_bas_u_i_q, grad_bas_u_j_q);

                const FLOAT *velocity_grad_quad_q = velocity_grad_quad + q*Dim*Dim;
                //const FLOAT *temprature_quad_q = temprature_quad + q;
                const FLOAT visc_quad_q = get_viscosity(velocity_grad_quad_q);

                for (k = 0; k < Dim; k++)
                for (l = 0; l < Dim; l++)
                {
                    F[j][l][i][k] += vol*w[q]*2*visc_quad_q*grad_bas_product[k+l*Dim]/1.0e5;
                    //printf("F: %e grad: %e w: %e vol: %e visc: %e\n", F[j][l][i][k], grad_bas_product[k+l*Dim], w[q], vol, visc);
                }
            }

            for (j = 0; j < N; j++)
            {
                const FLOAT *bas_p_j = phgQuadGetBasisValues(e, pressure, j, quad);
                for (k = 0; k < Dim; k++)
                {
                    B[j][i][k] -= vol*w[q]*bas_p_j[q]*grad_bas_u_i_q[k]/1.0e5;
                    Bt[i][k][j] -= vol*w[q]*bas_p_j[q]*grad_bas_u_i_q[k];
            
                }
            }
            }

            for (j = 0; j < N; j++)
            {
                const FLOAT *bas_p_j = phgQuadGetBasisValues(e, pressure, j, quad);
                C[j][j] += vol*w[q]*eps*bas_p_j[q]*bas_p_j[q];
            }
        
        }

#if 0
        for (s = 0; s < NFace; s++)
        {
            if (e->bound_type[s] & BC_BOT)
            {
                FLOAT lambda[Dim+1], lambda_x, lambda_y, lambda_z;
                INT quad_order = 5;
				QUAD *quad = phgQuadGetQuad2D(quad_order);
				FLOAT *w = quad->weights;
				FLOAT *p = quad->points;

				FLOAT area = phgGeomGetFaceArea(g, e, s);
				const FLOAT *normal = phgGeomGetFaceOutNormal(g, e, s);

                INT v0 = GetFaceVertex(s, 0);
                INT v1 = GetFaceVertex(s, 1);
                INT v2 = GetFaceVertex(s, 2);
                lambda[s] = 0;

				INT M_face = 3*(velocity->type->np_vert + velocity->type->np_edge);

                //printf("%d %d %d\n", M_face, velocity->type->np_vert, velocity->type->np_edge);

				SHORT bas_idx_e[M_face];

				phgDofGetBasesOnFace(velocity, e, s, bas_idx_e);

				for (q = 0; q < quad->npoints; q++)
				{
                    FLOAT vu[Dim];
                    lambda[v0] = *(p++);
                    lambda[v1] = *(p++);
                    lambda[v2] = *(p++);
                    
                    phgGeomLambda2XYZ(g, e, lambda, &lambda_x, &lambda_y, &lambda_z);

                    phgDofEval(velocity, e, lambda, vu);
                    
                    const FLOAT *bas = velocity->type->BasFuncs(velocity, e, 0, -1, lambda);
                    
                    for (i = 0; i < M_face; i++)
                    {
                        INT i_e = bas_idx_e[i];
                        for (j = 0; j < M_face; j++)
                        {
                            INT j_e = bas_idx_e[j];
                            for (k = 0; k < Dim; k++)
                            {
                                if (iter == 0)
                                F[j_e][k][i_e][k] += area*w[q]*bas[i_e]*bas[j_e]*beta_slip/1.0e5;
                                else
                                F[j_e][k][i_e][k] += area*w[q]*bas[i_e]*bas[j_e]*1e7/pow(sec_per_y, 1./3.)*pow(sqrt(vu[0]*vu[0]+vu[1]*vu[1]+vu[2]*vu[2]), 1./3.-1)/1.0e5;

                            }
                        }
                    }
                }
            }
        }
#endif
 


        for (s = 0; s < NFace; s++)
		{
			if ((e->bound_type[s] & BC_BOT_S))
			{
                FLOAT lambda[Dim+1], lambda_x, lambda_y, lambda_z;
                INT quad_order = 5;
                FLOAT Ns;
				QUAD *quad = phgQuadGetQuad2D(quad_order);
				FLOAT *w = quad->weights;
				FLOAT *p = quad->points;

				FLOAT area = phgGeomGetFaceArea(g, e, s);
				const FLOAT *normal = phgGeomGetFaceOutNormal(g, e, s);

                INT v0 = GetFaceVertex(s, 0);
                INT v1 = GetFaceVertex(s, 1);
                INT v2 = GetFaceVertex(s, 2);
                lambda[s] = 0;



                //if (fabs(normal[2]) < 1.0e-8)
                //{
                //    Ns = 1.0e50;
                //}
                //else
                    Ns = sqrt(1+SQUARE(normal[0]/normal[2])+SQUARE(normal[1]/normal[2]));

				FLOAT dt = 1.0;

				INT M_face = 3*(velocity->type->np_vert + velocity->type->np_edge);
				SHORT bas_idx_e[M_face];

				phgDofGetBasesOnFace(velocity, e, s, bas_idx_e);

				for (q = 0; q < quad->npoints; q++)
				{
                    lambda[v0] = *(p++);
                    lambda[v1] = *(p++);
                    lambda[v2] = *(p++);
                    
                    phgGeomLambda2XYZ(g, e, lambda, &lambda_x, &lambda_y, &lambda_z);
                    
                    const FLOAT *bas = velocity->type->BasFuncs(velocity, e, 0, -1, lambda, 0);
                    
                    for (i = 0; i < M_face; i++)
                    {
                        INT i_e = bas_idx_e[i];
                        const FLOAT *bas_i = phgQuadGetBasisValues(e, velocity, i_e, quad);
                        for (j = 0; j < M_face; j++)
                        {
                            INT j_e = bas_idx_e[j];
                            const FLOAT *bas_j = phgQuadGetBasisValues(e, velocity, j_e, quad);
                            
                            const FLOAT *bas_dot_n = get_bas_dot_normal(bas, normal, i_e, j_e);
                            for (k = 0; k < Dim; k++)
                            {
                                for (l = 0; l < Dim; l++)
                                {
                                    //printf("%f\n", bas_dot_n[k+l*Dim]);
                                    F[j_e][l][i_e][k] += area*w[q]*rho_w*gravity*bas_dot_n[k+l*Dim]*Ns*dt/1.0e5;
                                }
                            }
                        }
                    }
				}
			}
		}


        /*
        for (i = 0; i < M; i++)
        {
            INT i_d = phgDofMapE2D(trans, e, i*Dim*Dim);
            FLOAT *trans_data = DofData(trans);
            const FLOAT *trans_e = trans_data + i_d;
            trans_left(&F[i][0][0][0], Dim*M, Dim*M, trans_e);
            trans_rightT(&F[0][0][i][0], Dim*M, Dim*M, trans_e);
            trans_left(&Bt[i][0][0], N, N, trans_e);
            trans_rightT(&B[0][i][0], N, Dim*M, trans_e);
        }
        */


        for (i = 0; i < M; i++) {
            for (k = 0; k < Dim; k++) {
                if (phgDofDirichletBC_(velocity, e, i*Dim+k, NULL, bufu, &rhsu[i][0], DOF_PROJ_NONE))
                {
                    phgMatAddEntries(matF, 1, Iu[i] + k, M, Ju[k], bufu);
                }
                else
                {

                    phgMatAddEntries(matF, 1, Iu[i] + k, M*Dim, Iu[0],
                             &(F[i][k][0][0]));
                    phgMatAddEntries(matBt, 1, &Iu[i][k], N, Ip,
                             &Bt[i][k][0]);
                }
            }
        }
        
        
        for (i = 0; i < N; i++) 
        {
            phgMatAddEntries(matB, 1, Ip + i, M * Dim, Iu[0],
                          &B[i][0][0]);
            phgMatAddEntries(matC, 1, Ip + i, N, Ip,
                          &C[i][0]);
        } 



    }


}

void
build_rhs(SOLVER *solver, DOF **dofs, MAT **mats)
{
    DOF *velocity = dofs[0], *pressure = dofs[1];
    DOF *gradu;
    GRID *g = velocity->g;
    SIMPLEX *e;

    int M = velocity->type->nbas;	/* num of bases of Velocity */
    int N = pressure->type->nbas;	/* num of bases of Pressure */


    unpackDof(dofs, 3, &velocity, &pressure, &gradu);
    BTYPE DB_masks[3] = {BC_BOT|BC_HEA, BC_BOT|BC_LAT|BC_HEA, BC_BOT};
    phgDofSetDirichletBoundaryMasks(velocity, DB_masks);

    ForAllElements(g, e)
    {
        int i, j, k, q, s;
        int quad_order = 5;
        FLOAT rhs_u[M][Dim], rhs_p[N], bufu[M];
        INT local_map_idx_u[M][Dim], local_map_idx_p[N];
        QUAD *quad = phgQuadGetQuad3D(quad_order);
        FLOAT *w = quad->weights;
        FLOAT *p = quad->points;

        memset(rhs_u, 0, sizeof rhs_u); memset(rhs_p, 0, sizeof rhs_p);
        //calculate element volume

        //FLOAT *u_last_quad = phgQuadGetDofValues(e, u_last, quad);
        //FLOAT *p_last_quad = phgQuadGetDofValues(e, p_last, quad);
        //FLOAT *grad_u_last_quad = phgQuadGetDofValues(e, grad_u_last, quad);

        for (q = 0; q < quad->npoints; q++)
        {
            FLOAT vol = phgGeomGetVolume(g, e);
            //FLOAT *grad_u_last_quad_q = grad_u_last_quad + q*Dim*Dim;

            // calculate strain rate
            //FLOAT *strain_rate_last_quad_q = get_strain_rate(grad_u_last_quad_q);

            
            for (i = 0; i < M; i++)
            {
                const FLOAT *bas_u_i_q = phgQuadGetBasisValues(e, velocity, i, quad) + q;       
                for (k = 0; k < Dim; k++)
                {
                    if (k == 2){
                    //printf("%e\n", rhs_u[i][k]);
                    rhs_u[i][k] += -vol*w[q]*rho_i* gravity*bas_u_i_q[0]/1.0e5; 
                    }
                }
            }

            for (j = 0; j < N; j++)
            {
                //FLOAT bas_p_j = phgQuadGetBasisValues(e, pressure, j, quad);
                //FLOAT div_u_last_quad_q = grad_u_last_quad_q[0] + grad_u_last_quad_q[4] + grad_u_last_quad_q[8];
                rhs_p[j] += 0;
            }
        }

        for (s = 0; s < NFace; s++)
        {
            if ((e->bound_type[s] & BC_BOT_S) || (e->bound_type[s] & BC_TER))
            {
                FLOAT lambda[Dim+1], lambda_x, lambda_y, lambda_z;
                INT quad_order = 5;
                QUAD *quad = phgQuadGetQuad2D(quad_order);
                FLOAT *p = quad->points;
                FLOAT *w = quad->weights;
                FLOAT area = phgGeomGetFaceArea(g, e, s);
                const FLOAT *normal = phgGeomGetFaceOutNormal(g, e, s);

                INT M_face = 3*(velocity->type->np_vert + velocity->type->np_edge);
                SHORT bas_idx_e[M_face];

                phgDofGetBasesOnFace(velocity, e, s, bas_idx_e);
                
                INT v0 = GetFaceVertex(s, 0);
                INT v1 = GetFaceVertex(s, 1);
                INT v2 = GetFaceVertex(s, 2);
                lambda[s] = 0;

                for (q = 0; q < quad->npoints; q++)
                {
                    lambda[v0] = *(p++);
                    lambda[v1] = *(p++);
                    lambda[v2] = *(p++);

                    phgGeomLambda2XYZ(g, e, lambda, &lambda_x, &lambda_y, &lambda_z);
                    
                    const FLOAT *bas = velocity->type->BasFuncs(velocity, e, 0, -1, lambda);

                    for (i = 0; i < M_face; i++)
                    {
                        INT i_e = bas_idx_e[i];
                        const FLOAT *bas_i = phgQuadGetBasisValues(e, velocity, i_e, quad);
                        for (k = 0; k < Dim; k++)
                        {
                        	if (lambda_z < 0)
                                rhs_u[i_e][k] += area*w[q]*rho_w*gravity*lambda_z*bas[i_e]*normal[k]/1.0e5;
                        	else
                        		rhs_u[i_e][k] += 0;
                        }
                    }
                }
            }
        }

        /*
        for (s = 0; s < NFace; s++)
        {
            if ((e->bound_type[s] == BC_TER))
            {
                FLOAT lambda[Dim+1], lambda_x, lambda_y, lambda_z;
                INT quad_order = 5;
                QUAD *quad = phgQuadGetQuad2D(quad_order);
                FLOAT *p = quad->points;
                FLOAT *w = quad->weights;
                FLOAT area = phgGeomGetFaceArea(g, e, s);
                const FLOAT *normal = phgGeomGetFaceOutNormal(g, e, s);

                INT M_face = 3*(velocity->type->np_vert + velocity->type->np_edge);
                SHORT bas_idx_e[M_face];

                phgDofGetBasesOnFace(velocity, e, s, bas_idx_e);
                
                INT v0 = GetFaceVertex(s, 0);
                INT v1 = GetFaceVertex(s, 1);
                INT v2 = GetFaceVertex(s, 2);
                lambda[s] = 0;

                for (q = 0; q < quad->npoints; q++)
                {
                    lambda[v0] = *(p++);
                    lambda[v1] = *(p++);
                    lambda[v2] = *(p++);

                    phgGeomLambda2XYZ(g, e, lambda, &lambda_x, &lambda_y, &lambda_z);
                    
                    const FLOAT *bas = velocity->type->BasFuncs(velocity, e, 0, -1, lambda);

                    for (i = 0; i < M_face; i++)
                    {
                        INT i_e = bas_idx_e[i];
                        const FLOAT *bas_i = phgQuadGetBasisValues(e, velocity, i_e, quad);
                        for (k = 0; k < Dim; k++)
                        {
                        	if (lambda_z < 0)
                                rhs_u[i_e][k] += area*w[q]*rho_w*gravity*lambda_z*bas[i_e]*normal[k]/1.0e5;
                        	else
                        		rhs_u[i_e][k] += 0;
                        }
                    }
                }
            }
        }
        */

        /*
        for (i = 0; i < M; i++)
        {
            INT i_d = phgDofMapE2D(trans, e, i*Dim*Dim);
            FLOAT *trans_data = DofData(trans);
            const FLOAT *trans_e = trans_data + i_d;
            trans_left(&rhs_u[i][0], 1, 1, trans_e);
        }
        */

        for (i = 0; i < M; i++)
        for (k = 0; k < Dim; k++)
        {
            local_map_idx_u[i][k] = phgMapE2L(solver->rhs->map, 0, e, i*Dim+k);
            //printf("%d\n", local_map_idx_u[i][k]);
        }
        for (j = 0; j < N; j++)
            local_map_idx_p[j] = phgMapE2L(solver->rhs->map, 1, e, j);

        FLOAT tmp[Dim];
        for (i = 0; i < M; i++)
        for (k = 0; k < Dim; k++)
            if (phgDofDirichletBC_(velocity, e, i*Dim+k, NULL, bufu, tmp, DOF_PROJ_NONE))
                rhs_u[i][k] = 0;

        phgSolverAddRHSEntries(solver, M*Dim, local_map_idx_u[0], rhs_u[0]);
        phgSolverAddRHSEntries(solver, N, local_map_idx_p, rhs_p);
    }

}



int
main(int argc, char *argv[])
{
    GRID *g;
    SIMPLEX *e;
    DOF *gradu, *dofs[16];
    MAT *mats[16];
    SOLVER *solver;
    
    phgInit(&argc, &argv);
    g = phgNewGrid(-1);
    phgImport(g, "prism.neu", FALSE);
    build_ice_domain(g);
    phgBalanceGrid(g, 1.0, -1, NULL, 0.);
    
    printf("nvert nvert_global: %d %d %d\n", g->nvert, g->nvert_global, g->rank);

    phgExportVTK(g, "ice_domain.vtk", NULL);

    DOF *u = phgDofNew(g, DOF_P2, Dim, "u", DofInterpolation);
    DOF *p = phgDofNew(g, DOF_P1, 1, "p", DofInterpolation);
    gradu = phgDofNew(g, DOF_P2, Dim*Dim, "u", DofInterpolation);
    //phgDofSetDataByValue(gradu, 0);
	//phgDofGradient(u, &gradu, NULL, "grad u");
	//phgDofSetFunction(gradu, DofInterpolation);

    //BTYPE DB_masks[3] = {BC_BOT|BC_HEA, BC_BOT|BC_LAT|BC_HEA, BC_BOT};
    //phgDofSetDirichletBoundaryMasks(u, DB_masks);

	phgPrintf("    %d DOF (u:%d, p:%d), %d elements, %d submesh%s, load imbalance: %lg\n",
		  DofGetDataCountGlobal(u) + DofGetDataCountGlobal(p),
		  DofGetDataCountGlobal(u), DofGetDataCountGlobal(p),
		  g->nleaf_global, g->nprocs,
		  g->nprocs > 1 ? "es" : "", (double)g->lif);

    
    FLOAT time_start[3], time_end[3];
    phgGetTime(time_start);
	MAP *Vmap = phgMapCreate(u, NULL);
	MAP *Pmap = phgMapCreate(p, NULL);
    MAT *matNS, *matF, *matB, *matBt, *matC;
    MAT *pmat[4];

    INT iter_V = 0;

    while(1)    
    {

        phgPrintf("iter_V: %d\n", iter_V);

	    matF = phgMapCreateMat(Vmap, Vmap);
	    matB = phgMapCreateMat(Pmap, Vmap);
	    matBt = phgMapCreateMat(Vmap, Pmap);
	    matC = phgMapCreateMat(Pmap, Pmap);

	    matF->handle_bdry_eqns = TRUE;
		matC->handle_bdry_eqns = TRUE;
	    matB->handle_bdry_eqns = FALSE;
	    matBt->handle_bdry_eqns = FALSE;

	    pmat[0] = matF;
	    pmat[1] = matBt;
	    pmat[2] = matB;
	    pmat[3] = matC;


	    matNS = phgMatCreateBlockMatrix(g, 2, 2, pmat, NULL, NULL);
	    solver = phgSolverCreate(SOLVER_DEFAULT, u, p, NULL);
	    phgMatDestroy(&solver->mat);
	    solver->mat = matNS;
	    solver->rhs->mat = solver->mat;

	    packDof(dofs, 3, u, p, gradu);	
	    packMat(mats, 4, matF, matB, matBt, matC);

	    phgPrintf("\tBuild rhs: ");
	    build_rhs(solver, dofs, mats);
	    phgPrintf("\tBuild matrices: ");
		build_matrices(solver, dofs, mats);
	    phgPrintf("\tAssemble linear system: ");
	    phgSolverAssemble(solver);

	    phgPrintf("\tSolution: ");
	    phgSolverSolve(solver, TRUE, u, p, NULL);

	    phgDofGradient(u, &gradu, NULL, "grad u");

        phgExportVTK(g, "ice.vtk", u, p, NULL);



        if (iter_V > 1)
        {
            break;
        }


        iter_V++;
    }



    phgGetTime(time_end);
    phgPrintf("elapsed_time: %f\n", time_end[1]-time_start[1]);

    phgDofFree(&gradu);
    phgDofFree(&p);
    phgDofFree(&u);
    phgFinalize();
    return 0;
}
