#include "phg.h"
#include <string.h>
#include <math.h>

static int
my_bc_map(int bctype)
{
    switch (bctype) {
    case 0:
        return DIRICHLET;
    case 1:
        return NEUMANN;
    case 2:
        return BDRY_USER3;
    case 3:
        return BDRY_USER4;
    case 4:
        return BDRY_USER5;
    case 5:
        return BDRY_USER6;
    default:
        return BDRY_USER7;
    }
}

int
main(int argc, char *argv[])
{
    char *fn = "../test/cube4.dat";
	char *file_map = NULL;
    GRID *g;
    int i, refines = 0;

    phgOptionsRegisterFilename("mesh_file", "Mesh file", &fn);
    phgOptionsRegisterFilename("map_file", "Map file", &file_map);
    phgOptionsRegisterInt("refines", "Refine times", &refines);
    phgInit(&argc, &argv);
    g = phgNewGrid(-1);
    phgImportSetBdryMapFunc(my_bc_map);
    if (!phgImport(g, fn, FALSE))
	phgError(1, "can't read file \"%s\".\n", fn);
    phgCheckConformity(g);
    phgPrintf("=====Initial mesh:%d\n", refines);
    phgPrintf("   %d elements, %d submeshes, load imbalance: %lg\n",
	      g->nleaf_global, g->nprocs,
	      (double)g->lif);

	if (file_map != NULL) {
		phgPrintf("open map file: %s\n", file_map);
	FILE *fp = fopen(file_map, "r");
	int idx;
	for (i = 0; i < g->nelem; i++) {
		fscanf(fp, "%d", &idx);
		g->roots[i].region_mark = idx;
	}
	fclose(fp);
	}


    phgExportMedit(g, "mixed.mesh");
    phgExportALBERT(g, "mixed.dat");
    phgExportVTK(g, "mixed.vtk", NULL);
    //phgExportTecplot(g, "mixed.plt", NULL);
    //phgExportEnsightT(g, "mixed", 1, 0, NULL);

#if 1				/* check distribute */
    if (phgBalanceGrid(g, 1.2, 1, NULL, 0.))
	phgPrintf("Repartition mesh, load imbalance: %lg\n",
		  (double)g->lif);
    phgCheckConformity(g);

    for (i = 0; i < refines; i++) {
	phgPrintf("=====Refine pass:%d\n", i);
	phgRefineAllElements(g, 1);
	phgPrintf("   %d elements, %d submeshes, load imbalance: %lg\n",
		  g->nleaf_global, g->nprocs,
		  (double)g->lif);
	if (phgBalanceGrid(g, 1.2, 1, NULL, 0.))
	    phgPrintf("Repartition mesh, load imbalance: %lg\n",
		      (double)g->lif);
	phgCheckConformity(g);
    }
#endif

    //phgFreeGrid(&g);
    phgFinalize();
    return 0;
}
