/* Tecplot output functions
 *
 * $Id: tecplot.c,v 1.3 2009/01/11 15:06:21 wleng Exp $
 */
// (highlight-regexp "phgIO\w+" "hi-red-b")
// (highlight-regexp "phgIO\w+" "hi-blue-b")
#define USE_VECTORS	1

#include "phg.h"
#include "io.h"
#include "periodic.h"

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>	/* PATH_MAX */
#include <ctype.h>


#define TECPLOT_FLOAT float
#define TECPLOT_DOUBLE double
#define TECPLOT_INT int
#define TECPLOT_SHORT_INT short int

#define phgIOWriteRoot1Value(g, x, Type)  {	\
	Type _tmp = (Type) (x);			\
	phgIOWriteRoot(g, (void *) &_tmp, sizeof(Type), 1);	\
    }

/********************/
/* Tecplot GLOBLA.h */
/********************/
#define TecplotBinaryFileVersion 111 /* copied from GLOBAL.h */

#define ZONEMARKER           299.0
#define GEOMMARKER           399.0
#define TEXTMARKER           499.0
#define CUSTOMLABELMARKER    599.0
#define USERRECMARKER        699.0
#define DATASETAUXDATAMARKER 799.0
#define VARAUXDATAMARKER     899.0
#define EOHMARKER            357.0

#define BAD_SET_VALUE        (-1)
#define STRAND_ID_STATIC     (-1)
#define STRAND_ID_PENDING    (-2)
#define STRAND_ID_UNASSIGNED (-3) /* ...only relevant for preplot */

#define   LARGEDOUBLE      1.0E+150
#define   SMALLDOUBLE      1.0E-150
#define   LARGELONG        2147483646
#define   LARGESHORT       32766
#define   DEGPERRADIANS    57.295779513082323

typedef enum FileType {FULLFILE,
                       GRIDFILE,
                       SOLUTIONFILE} FileType;
typedef enum {
    FieldDataType_Reserved, /* never use */
    FieldDataType_Float,
    FieldDataType_Double,
    FieldDataType_Int32,
    FieldDataType_Int16,
    FieldDataType_Byte,
    FieldDataType_Bit
} FieldDataType_e;

typedef enum ZoneDataType {ORDERED,
                           FELINESEG,
                           FETRIANGLE,
                           FEQUADRILATERAL,
                           FETETRAHEDRON,
                           FEBRICK,
                           FEPOLYGON,
                           FEPOLYHEDRON} ZoneDataType;


#ifdef MAX
#  undef MAX
#endif
#define MAX(X,Y) ((X) > (Y) ? (X) : (Y) )
#ifdef MIN
#  undef MIN
#endif
#define MIN(X,Y) ((X) < (Y) ? (X) : (Y) )

////////////////////////////////////////////////////////////////////////////////




/* buffer for broadcasting filename */
static const char *filename;
/* buffer for broadcasting list of DOF ids */
static USHORT *dof_ids = NULL;	/* last element for filename length */
static USHORT dof_ids_len = 0;

static unsigned char *buffer = NULL;
static size_t buffer_pos, buffer_size;
static BTYPE *types_vert = NULL, *types_swap;
#define SWAP_TYPES_VERT {			\
    types_swap = g->types_vert;			\
    g->types_vert = types_vert;			\
    types_vert = types_swap;			\
}

static void phgIODumpString(GRID *g, char *S)
{
    INT L = 0;

    while (S[L] != '\0') {
	phgIOWriteRoot1Value(g, (unsigned char)S[L++], TECPLOT_INT);
    }
    phgIOWriteRoot1Value(g, 0, INT);
}

#define AppendData(buf, size, count) {			\
	size_t size0 = (size) * (count);		\
	assert(buffer_pos + size0 <= buffer_size);	\
	memcpy(buffer + buffer_pos, buf, size0);	\
	buffer_pos += size0;				\
    }

/* no: dof num
 * m:  m-th componet dof */
static char *
parse_name(GRID *g, int no, int m)
{
    static char name0[128 + 16];
    char name[128];
    char *p = name;
    const char *s;
    DOF *dof = phgDofIndex2Pointer(g, dof_ids[no + 1]);
    int dim = dof->dim;
    int i, n;

    s = dof->name;
    assert(m < dim);

    /* check for duplicate names */
    for (i = 0, n = 0; i < no; i++)
	if (!strcmp(s, phgDofIndex2Pointer(g, dof_ids[i + 1])->name))
	    n++;

    while (*s != '\0' && p - name < sizeof(name) - 1) {
	*p = isspace(*(const BYTE *)s) ? '_' : *s;
	p++;
	s++;
    }
    *p = '\0';

    if (n == 0)
	sprintf(name0, "%s", name);
    else
	sprintf(name0, "%s_%d", name, n);

    if (dim == 1) 
	return name0;

    /* add [m] after dof name */
    strcpy(name, name0);
    sprintf(name0, "%s[%d]", name, m);
    
    return name0;
}


void
phgExportTecplot_(GRID *g)
/* Output grid in Tecplot unstructured grid format. */
{
    int k, a, dim, imax, imin;
    size_t size;
    INT i;
    INT nleaf, nvert;
    SIMPLEX *s;
    DOF *dof;
    FLOAT *values;
    static char fn[PATH_MAX + 1];
    char VerBuf[5];
    TECPLOT_INT NumVars = 3 + 2; /* with coordinate, region mark
				  * and submesh indices */
    double max_xyz[3], min_xyz[3];

    ParallelFunction(phgExportTecplot_, TRUE);
    if (g == NULL || g->nprocs <= phgRank)
	return;

    nleaf = g->nleaf;
    nvert = g->nvert;

    assert((g->flags & VERT_FLAG));
    assert(g->rank != 0 || sizeof(a) == 4);	/* to be fixed later */


    /**********************/
    /* Broadcast dofs     */
    /**********************/
#if USE_MPI
    if (phgMasterSlave) {
	if (g->rank == 0) {
	    memcpy(fn, filename, a = strlen(filename) + 1);
	    dof_ids[dof_ids_len - 1] = a ;
	}
	filename = fn;
	MPI_Bcast(&dof_ids_len, sizeof(dof_ids_len), MPI_BYTE, 0, g->comm);
	MPI_Bcast(dof_ids, sizeof(*dof_ids)*dof_ids_len, MPI_BYTE, 0, g->comm);
	a = dof_ids[dof_ids_len - 1];
	MPI_Bcast((void *)filename, a, MPI_BYTE, 0, g->comm);
    }
    else
#else
	{
	    strcpy(fn, filename);
	    filename = fn;
	}
#endif


    /************/
    /* Openfile */
    /************/
    filename = phgIOAddExtension(filename, "plt");
    if (!phgIOOpen(g, (char *)filename)) {
	phgError(0, "cannot create Tecplot file \"%s\".\n", filename);
	return;
    }
    if (g->rank == 0) phgInfo(1, "creating Tecplot file \"%s\".\n", filename);

    buffer_size = nvert * sizeof(TECPLOT_FLOAT) * 3;
    if ((size = nleaf * 5 * sizeof(a)) > buffer_size)
	buffer_size = size;

    /********************************/
    /* compute buf size of DOF data */
    /********************************/
    for (a = 0; a < dof_ids[0]; a++) {
	dof = phgDofIndex2Pointer(g, dof_ids[a + 1]);
        dim = DofDim(dof);
	dim = dim > 4 ? 4 : dim;
	if (dof->type == DOF_P0 || dof->type == DOF_DG0) {
	    size = nleaf * dim * sizeof(TECPLOT_FLOAT);
	}
	else {
	    size = nvert * dim * sizeof(TECPLOT_FLOAT);
	}
	if (buffer_size < size)
	    buffer_size = size;
    }
    buffer = phgAlloc(buffer_size);


    /****************************/
    /* File header, dof info... */
    /****************************/
    // Write out header information

    sprintf(VerBuf, "V%-3d", TecplotBinaryFileVersion);
    phgIOPrintf(g, "#!TD%s", VerBuf);

    // byte order
    phgIOWriteRoot1Value(g, 1, TECPLOT_INT);

    // Write out filetype
    phgIOWriteRoot1Value(g, FULLFILE, TECPLOT_INT);
    
    // Write title
    phgIODumpString(g, "Example: FE-Volume Tetrahedral Data");

    // Number of variables (NumVar) in the datafile
    for (a = 0; a < dof_ids[0]; a++) {
	dof = phgDofIndex2Pointer(g, dof_ids[a + 1]);
	NumVars += dof->dim; 
    }
    phgIOWriteRoot1Value(g, NumVars, TECPLOT_INT);
	
    // Variable names
    phgIODumpString(g, "X");
    phgIODumpString(g, "Y");
    phgIODumpString(g, "Z");
    for (a = 0; a < dof_ids[0]; a++) {
	dof = phgDofIndex2Pointer(g, dof_ids[a + 1]);
	dim = dof->dim;
	for (k = 0; k < dim; k++)
	    phgIODumpString(g, parse_name(g, a, k));
    }
    phgIODumpString(g, "region_mark");
    phgIODumpString(g, "submesh_no");

    // Zone marker. Value = 299.0
    phgIOWriteRoot1Value(g, ZONEMARKER, TECPLOT_FLOAT);

    // Zone name
    phgIODumpString(g, "ZONE 001");

    // ParentZone: no ParentZone (zero based)
    phgIOWriteRoot1Value(g, BAD_SET_VALUE, TECPLOT_INT);

    // StrandID: static StrandID (zero based)
    phgIOWriteRoot1Value(g, STRAND_ID_STATIC, TECPLOT_INT);

    // Solution time: zero time
    phgIOWriteRoot1Value(g, 0.0, TECPLOT_DOUBLE);

    // Zone Color: No Color
    phgIOWriteRoot1Value(g, -1, TECPLOT_INT);

    // Zone type: FETETRAHEDRON
    phgIOWriteRoot1Value(g, FETETRAHEDRON, TECPLOT_INT);
    
    // DataPacking 0=Block
    phgIOWriteRoot1Value(g, 0, TECPLOT_INT);

    // Specify Var Location: true
    phgIOWriteRoot1Value(g, 1, TECPLOT_INT);

    // Variable Location: 0 = Node, 1 = Cell Centered
    phgIOWriteRoot1Value(g, 0, TECPLOT_INT); /* x */
    phgIOWriteRoot1Value(g, 0, TECPLOT_INT); /* y */
    phgIOWriteRoot1Value(g, 0, TECPLOT_INT); /* z */
    for (a = 0; a < dof_ids[0]; a++) {
	dof = phgDofIndex2Pointer(g, dof_ids[a + 1]);
	if (dof->type == DOF_P0 || dof->type == DOF_DG0) {
	    for (k = 0; k < dof->dim; k++)
		phgIOWriteRoot1Value(g, 1, TECPLOT_INT);
	} else {
	    for (k = 0; k < dof->dim; k++)
		phgIOWriteRoot1Value(g, 0, TECPLOT_INT);
	}
    }
    phgIOWriteRoot1Value(g, 1, TECPLOT_INT); /* region mark */
    phgIOWriteRoot1Value(g, 1, TECPLOT_INT); /* submesh no */

    // Are raw face neighbors available? No
    phgIOWriteRoot1Value(g, 0, TECPLOT_INT);
    
    // Num User defined face neighbors: 0
    phgIOWriteRoot1Value(g, 0, TECPLOT_INT);

    // NumPts
    phgIOWriteRoot1Value(g, g->nvert_global, TECPLOT_INT);

    // NumElements
    phgIOWriteRoot1Value(g, g->nelem_global, TECPLOT_INT);

    // Usused ICellDim,JCellDim,KCellDim 
    phgIOWriteRoot1Value(g, 0, TECPLOT_INT);
    phgIOWriteRoot1Value(g, 0, TECPLOT_INT);
    phgIOWriteRoot1Value(g, 0, TECPLOT_INT);

    // No more aux data name/value pairs.
    phgIOWriteRoot1Value(g, 0, TECPLOT_INT);

    // separate the header from the data with an EOHMARKER
    phgIOWriteRoot1Value(g, EOHMARKER, TECPLOT_FLOAT);

    // Zone marker  Value = 299.0
    phgIOWriteRoot1Value(g, ZONEMARKER, TECPLOT_FLOAT);

    // Variable data format for all vars
    phgIOWriteRoot1Value(g, FieldDataType_Float, TECPLOT_INT); /* x */
    phgIOWriteRoot1Value(g, FieldDataType_Float, TECPLOT_INT); /* y */
    phgIOWriteRoot1Value(g, FieldDataType_Float, TECPLOT_INT); /* z */
    for (a = 0; a < dof_ids[0]; a++) {
	dof = phgDofIndex2Pointer(g, dof_ids[a + 1]);
	for (k = 0; k < dof->dim; k++)
	    phgIOWriteRoot1Value(g, FieldDataType_Float, TECPLOT_INT);
    }
    phgIOWriteRoot1Value(g, FieldDataType_Float, TECPLOT_INT); /* region mark */
    phgIOWriteRoot1Value(g, FieldDataType_Float, TECPLOT_INT); /* submesh no */

    // Has passive variables: no,
    phgIOWriteRoot1Value(g, 0, TECPLOT_INT);

    // Has passive variables: no,
    phgIOWriteRoot1Value(g, 0, TECPLOT_INT);

    // Zero based zone number to share connectivity
    // list with: no sharing
    phgIOWriteRoot1Value(g, -1, TECPLOT_INT);

    /**************************/
    /* Variable Min Max value */
    /**************************/
    /* Todo: Look up for phg limit. */

    /* XYZ Range*/
    min_xyz[0] = 1e20; max_xyz[0] = -1e20;
    min_xyz[1] = 1e20; max_xyz[1] = -1e20;
    min_xyz[2] = 1e20; max_xyz[2] = -1e20;
    for (i = 0; i < nvert; i++) {
	for (a = 0; a < 3; a++) {
	    min_xyz[a] = MIN(min_xyz[a], g->verts[i][a]);
	    max_xyz[a] = MAX(max_xyz[a], g->verts[i][a]);
	}
    }
#if USE_MPI
    {
	double min_xyz_global[3], max_xyz_global[3];
	MPI_Reduce(min_xyz, min_xyz_global, 3, 
		   MPI_DOUBLE, MPI_MIN, 0, g->comm);
	MPI_Reduce(max_xyz, max_xyz_global, 3, 
		   MPI_DOUBLE, MPI_MAX, 0, g->comm);
	memcpy(min_xyz, min_xyz_global, 3*sizeof(*min_xyz));
	memcpy(max_xyz, max_xyz_global, 3*sizeof(*max_xyz));
    }
#endif
    for (a = 0; a < 3; a++) {
	phgIOWriteRoot1Value(g, min_xyz[a], TECPLOT_DOUBLE);
	phgIOWriteRoot1Value(g, max_xyz[a], TECPLOT_DOUBLE);
	//printf("(%f, %f) ", min_xyz[a], max_xyz[a]);
    }

    /* Varialbe Range */
    for (a = 0; a < dof_ids[0]; a++) {
	dof = phgDofIndex2Pointer(g, dof_ids[a + 1]);
	dim = dof->dim;
	if (dof->type == DOF_P0 || dof->type == DOF_DG0) {
	    FLOAT min[dim], max[dim];
	    for (k = 0; k < dim; k++) {
		min[k] = 1e20;
		max[k] = -1e20;
	    }

	    ForAllElements(g, s) {
		values = DofElementData(dof, s->index);
		for (k = 0; k < dim; k++) {
		    min[k] = MIN(min[k], values[k]);
		    max[k] = MAX(max[k], values[k]);
		}
	    }
#if USE_MPI
	    {
		double min_global[dim], max_global[dim];
		MPI_Reduce(min, min_global, dim, 
			   MPI_DOUBLE, MPI_MIN, 0, g->comm);
		MPI_Reduce(max, max_global, dim, 
			   MPI_DOUBLE, MPI_MAX, 0, g->comm);
		memcpy(min, min_global, dim*sizeof(*max));
		memcpy(max, max_global, dim*sizeof(*max));
	    }
#endif
	    for (k = 0; k < dim; k++) {
		phgIOWriteRoot1Value(g, min[k], TECPLOT_DOUBLE);
		phgIOWriteRoot1Value(g, max[k], TECPLOT_DOUBLE);
	    }
	} else {
	    FLOAT min[dim], max[dim];
	    for (k = 0; k < dim; k++) {
		min[k] = 1e20;
		max[k] = -1e20;
	    }

	    values = DofVertexData(dof, 0);
	    for (i = 0; i < nvert; i++) {
		for (k = 0; k < dim; k++) {
		    min[k] = MIN(min[k], values[k]);
		    max[k] = MAX(max[k], values[k]);
		}
		values += dim;
	    }
#if USE_MPI
	    {
		double min_global[dim], max_global[dim];
		MPI_Reduce(min, min_global, dim, 
			   MPI_DOUBLE, MPI_MIN, 0, g->comm);
		MPI_Reduce(max, max_global, dim, 
			   MPI_DOUBLE, MPI_MAX, 0, g->comm);
		memcpy(min, min_global, dim*sizeof(*max));
		memcpy(max, max_global, dim*sizeof(*max));
	    }
#endif
	    for (k = 0; k < dim; k++) {
		phgIOWriteRoot1Value(g, min[k], TECPLOT_DOUBLE);
		phgIOWriteRoot1Value(g, max[k], TECPLOT_DOUBLE);
	    }
	}
    } /* end of Variable range */

    /* region mark range */
    imin = 100000; imax = -10000; 
    ForAllElements(g, s) {
	imin = MIN(imin, s->region_mark);
	imax = MAX(imax, s->region_mark);
    }
#if USE_MPI
    {
	int imin_global, imax_global;
	MPI_Reduce(&imin, &imin_global, 1, 
		   MPI_INT, MPI_MIN, 0, g->comm);
	MPI_Reduce(&imax, &imax_global, 1, 
		   MPI_INT, MPI_MAX, 0, g->comm);
	imin = imin_global;
	imax = imax_global;
    }
#endif
    phgIOWriteRoot1Value(g, imin, TECPLOT_DOUBLE);
    phgIOWriteRoot1Value(g, imax, TECPLOT_DOUBLE);
    
    /* submesh indices range */
    imin = 100000; imax = -10000; 
    ForAllElements(g, s) {
	imin = MIN(imin, g->rank);
	imax = MAX(imax, g->rank);
    }
#if USE_MPI
    {
	int imin_global, imax_global;
	MPI_Reduce(&imin, &imin_global, 1, 
		   MPI_INT, MPI_MIN, 0, g->comm);
	MPI_Reduce(&imax, &imax_global, 1, 
		   MPI_INT, MPI_MAX, 0, g->comm);
	imin = imin_global;
	imax = imax_global;
    }
#endif
    phgIOWriteRoot1Value(g, imin, TECPLOT_DOUBLE);
    phgIOWriteRoot1Value(g, imax, TECPLOT_DOUBLE);


    /*************/
    /* Vertices  */
    /*************/
    for (a = 0; a < Dim; a++) {
	buffer_pos = 0;
	for (i = 0; i < nvert; i++) {
	    TECPLOT_FLOAT x = g->verts[i][a];
	    AppendData(&x, sizeof(x), 1);
	}
#if USE_MPI
	phgIOWrite(g, buffer, sizeof(TECPLOT_FLOAT), nvert, g->nvert_global,
		   g->types_vert, g->L2Gmap_vert, TRUE);
#else
	phgIOWrite(g, buffer, sizeof(TECPLOT_FLOAT), nvert, g->nvert_global,
		   NULL, NULL, FALSE);
#endif
    }

    /************/
    /* Dof data */
    /************/
    for (a = 0; a < dof_ids[0]; a++) {
	dof = phgDofIndex2Pointer(g, dof_ids[a + 1]);

	if (DofIsHP(dof)) {
	    phgError(0, "Unimplement output tecplot for HP Dof.\n");
	}

	if (g->rank == 0)
	    phgInfo(2, "writing DOF %d: \"%s\"\n", a, dof->name);
	dim = dof->dim;

	if (dof->type == DOF_P0 || dof->type == DOF_DG0) {
	    /* Cell Data */
	    for (k = 0; k < dim; k++) {
		buffer_pos = 0;
		ForAllElements(g, s) {
		    values = DofElementData(dof, s->index);
		    TECPLOT_FLOAT v = values[k];
		    AppendData(&v, sizeof(v), 1);
		}
		phgIOWrite(g, buffer, sizeof(TECPLOT_FLOAT), nleaf,
			   g->nelem_global, NULL, NULL, FALSE); 
	    }
	} else {
	    /* Vertex Date */
	    DOF *dof0;
	    dof = dof0 = phgDofIndex2Pointer(g, dof_ids[a + 1]);
	    if (dof->type != DOF_P1) {
		SWAP_TYPES_VERT;
		dof = phgDofCopy(dof0, NULL, DOF_P1, dof0->name);
		SWAP_TYPES_VERT;
	    }

	    for (k = 0; k < dim; k++) {
		buffer_pos = 0;
		values = DofVertexData(dof, 0) + k;
		for (i = 0; i < nvert; i++) {
		    TECPLOT_FLOAT v = *values;
		    AppendData(&v, sizeof(v), 1);
		    values += dim;
		}
#if USE_MPI
		phgIOWrite(g, buffer, sizeof(TECPLOT_FLOAT), nvert,
			   g->nvert_global, g->types_vert, g->L2Gmap_vert, TRUE); 
#else
		phgIOWrite(g, buffer, sizeof(TECPLOT_FLOAT), nvert,
			   g->nvert_global, NULL, NULL, FALSE); 
#endif
	    }

	    if (dof != dof0)
		phgDofFree(&dof);
	}
    } /* end of variables */


    /***********************************/
    /* save regional mark as cell data */
    /***********************************/
    buffer_pos = 0;
    ForAllElements(g, s) {
	TECPLOT_FLOAT v = s->region_mark;
	AppendData(&v, sizeof(v), 1);
    }
    phgIOWrite(g, buffer, sizeof(TECPLOT_FLOAT), nleaf, g->nelem_global, NULL, NULL, FALSE);

    /*************************************/
    /* save submesh indices as cell data */
    /*************************************/
    buffer_pos = 0;
    ForAllElements(g, s) {
	TECPLOT_FLOAT v = g->rank;
	AppendData(&v, sizeof(v), 1);
    }
    phgIOWrite(g, buffer, sizeof(TECPLOT_FLOAT), nleaf, g->nelem_global, NULL, NULL, FALSE);

    /**************/
    /* tetrahedra */
    /**************/
    buffer_pos = 0;
    /* element vertices */
    ForAllElements(g, s) {
	TECPLOT_INT a = 4;
	for (i = 0; i < 4; i++) {
	    a = GlobalVertex(g, s->verts[i]);
	    AppendData(&a, sizeof(a), 1);
	}
    }
    phgIOWrite(g, buffer, sizeof(TECPLOT_INT) * 4, nleaf, g->nelem_global, NULL, NULL,
	       FALSE); 

    phgIOClose();

    phgFree(buffer);
    buffer = NULL;

    return;
}


const char *
phgExportTecplotn(GRID *g, const char *fn, int ndof, DOF **dofs)
/* Outputs mesh in Tecplot format, returns the output filename.
 * The variable part of the arguments is a NULL terminated list of DOFs
 * to save to the Tecplot file */
{
    int i, j, k;
    DOF *dof;

    FreeAtExit(dof_ids);
    types_vert = NULL;

    assert(fn != NULL && Dim == 3);
    if (g == NULL)
	return NULL;

    /* output fn */
    filename = fn;
    /* initialize dof_ids */
    dof_ids_len = ndof + 2;
    phgFree(dof_ids);
    dof_ids = phgAlloc(sizeof(*dof_ids) * dof_ids_len);

    /* list of DOFs */
    for (i = 0, k = 0; i < ndof; i++) {
	if ((dof = dofs[i])->type == DOF_ANALYTIC) {
	    if (g->rank == 0) {
		phgWarning("DOF type 'DOF_ANALYTIC' is not supported by "
			   "phgExportTecplot(). \n");
		phgWarning("DOF '%s' will not be saved.\n", dof->name);
	    }
	    continue;
	}
	if (dof->g != g)
	    phgError(1, "%s: DOF \"%s\" doesn't belong to the grid.\n",
		     __func__, dof->name);
	/* test for missing NULL terminator */
	if (!phgDofIsValid(g, dof)) {
	    if (g->rank == 0)
		phgWarning("phgExportTecplot: missing NULL terminator"
			   " in argument list.\n");
	    break;
	}
	if ((j = phgDofPointer2Index(dof)) < 0)
	    break;
	dof_ids[++k] = j;
    }
    dof_ids[0] = k;

    if (g->period != NULL) {
	/* reconstruct types_vert[] without periodicity */
	BYTE	flags;
	PERIOD	*period;
	INT	*owner_index_vert;
	SHORT	*owner_rank_vert;

	period = g->period;
	g->period = NULL;

	/* save */
	types_vert = g->types_vert;
	g->types_vert = NULL;
	owner_index_vert = g->owner_index_vert;
	g->owner_index_vert = NULL;
	owner_rank_vert = g->owner_rank_vert;
	g->owner_rank_vert = NULL;
	flags = g->flags;
	g->flags = VERT_FLAG;

	/* reconstruct types_vert, etc */
	phgUpdateBoundaryTypes(g);

	/* restore, except types_vert */
	g->period = period;
	phgFree(g->owner_index_vert);
	g->owner_index_vert = owner_index_vert;
	phgFree(g->owner_rank_vert);
	g->owner_rank_vert = owner_rank_vert;
	g->flags = flags;
    }

    phgExportTecplot_(g);

    if (g->period != NULL) {
	/* restore types_vert */
	phgFree(g->types_vert);
	g->types_vert = types_vert;
    }

    return filename;
}

const char *
phgExportTecplot(GRID *g, const char *fn, DOF *dof, ...)
{
    int ndof;
    DOF **dofs = phgAlloc(256 * sizeof(*dofs));
    va_list ap;

    va_start(ap, dof);
    for (ndof = 0; ndof < 256; ndof++) {
	if (dof == NULL)
	    break;
	dofs[ndof] = dof;
	dof = va_arg(ap, DOF *);
    }
    va_end(ap);

    fn = phgExportTecplotn(g, fn, ndof, dofs);
    phgFree(dofs);

    return fn;
}
